#!/usr/bin/wish
#-------------------------------------------------------------------------------
# CharForm - uses a form-like interface to generate a character sheet in HTML.
#-------------------------------------------------------------------------------
# Copyright (C) 1997 by James MacKay
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# The author can be contacted via e-mail at: jmackay@sentex.net
#--------------------------------------------------------------------------
# Applet Parameters:
#
#   numPowers - the number of lines on which powers (skills, etc.) can
#               be entered (default = 15)
#   numDisads - the number of lines on which disadvantages can be
#               entered (default = 10)
#   numNotes  - the number of notes that can be entered (default = 5)
#   precision - the number of significant digits in floating point values
#               used in the program (e.g. in calculators)
#--------------------------------------------------------------------------
# Global Variables:
# ================
#
# Characteristics:
# ---------------
# CSchar(count)        - the total number of characteristics
# CSchar(totalCost)    - the total cost of all characteristics
# CSchar($i,name)      - the name of the $i`th characteristic (e.g. STR)
# CSchar($i,value)     - the value of the $i`th characteristic **
# CSchar($i,cost)      - the cost of the $i`th characteristic **
# CSchar(dataType)     - the type of all input characteristics data ++
# 
# Powers:
# ------
# CSpow(count)     - the total number of powers that can be entered
# CSpow(totalCost) - the total cost of all powers
# CSpow($i,cost)   - the cost of the $i`th power **
# CSpow($i,desc)   - the description of the $i`th power
# CSpow($i,end)    - the endurance cost of the $i`th power
# CSpow(dataType)  - the type of all input powers data ++
#
# Disadvantages:
# -------------
# CSdisad(count)     - the total number of disadvantages that can be entered  
# CSdisad(totalCost) - the total cost/value of all disadvantages
# CSdisad($i,cost)   - the cost/value of the $i`th disadvantage **
# CSdisad($i,desc)   - the description of the $i`th disadvantage
# CSdisad(dataType)  - the type of all input disadvantages data ++
#
# Notes:
# -----
# CSnote(count)    - the total number of notes that can be entered
# CSnote($i,desc)  - the contents of the $i`th note
# CSnote(dataType) - the type of all input notes data ++
#
# Character Information:
# ---------------------
# CSinfo(name)        - the character`s name
# CSinfo(name,type)   - the type of all input name data ++
# CSinfo(creator)     - the name of the character`s creator
# CSinfo(comment)     - the creator`s comment
# CSinfo(imageURL)    - the URL of the (main) picture of the character
# CSinfo(origin,text) - the text widget containing the character`s origin
# CSinfo(origin,type) - the type of all input origin data ++
#
# Output and Clear:
# ----------------
# CSout(type)      - how the character sheet is to be output %%
# CSout(text)      - the text widget to which the character sheet can be output
# CSout(buffer)    - holds the data to be output
# CSout(clear)     - 1 if clearing all character sheet info has been confirmed,
#                    0 if it hasn`t
# CSout(addresses) - the comma-separated list of e-mail addresses to which the
#                    character sheet is to be sent
# CSout(filename)  - the name of the file the character sheet is to be output to
# CSout(fileID)    - the fileID to which the character sheet is to be output
# CSout(browserPolicyAccessed) - 1 if we have access to the browser security
#                                policy, and 0 if we don`t
# CSout(result)    - description of the result of the last attempted output
#
# Calculators:
# -----------
# CScalc($ID,q) - the calculation associated with the calculator uniquely
#                 identified by $ID
# CScalc($ID,a) - the answer to the last performed calculation done by the
#                 calculator uniquely identified by $ID
#
# Pages (for Pager):
# -----------------
# CSpage(count)       - the number of pages in the pager
# CSpage($i,title)    - the title of the (i+1)th page
# CSpage($i,page)     - the top frame for the (i+1)th page
# CSpage(currIndex)   - the index of the current page
# CSpage(currTitle)   - the title of the current page
# CSpage(currPageNum) - the page number label text for the current page
# CSpage(pageHolder)  - the frame that holds the pages
#
#   ** - these include any non-digit characters
#   ++ - current data types are 'text' and 'html'
#   %% - current output types are: 'text', 'email', 'file' and 'browse'
#--------------------------------------------------------------------------

#-------------------------------------------------------------------------------
# Configure the application.
#-------------------------------------------------------------------------------
proc configure {} {
   global CSchar CSpow CSdisad CSnote CSinfo CSout \
          embed_args tcl_precision

   # set characteristics UI constants:
   set charNames [list STR DEX CON BODY INT EGO PRE COM PD ED SPD REC END STUN]
   set CSchar(count) [llength $charNames]
   for {set i 0} {$i < $CSchar(count)} {incr i} {
      set CSchar($i,name)  [lindex $charNames $i]
      }

   # get the number of powers, notes and disadvantages:
   set CSpow(count) \
      [expr {[info exists embed_args(numPowers)] ? $embed_args(numPowers) : 15}]
   set CSdisad(count) \
      [expr {[info exists embed_args(numDisads)] ? $embed_args(numDisads) : 10}] 
   set CSnote(count) \
      [expr {[info exists embed_args(numNotes)] ? $embed_args(numNotes) : 5}]

   # set the precision (for all of the calculators):
#   set tcl_precision \
#      [expr {[info exists embed_args(precision)] ? $embed_args(precision) : 8}]

   # (re)set all of the information in the character sheet:
   resetAllInfo

   }  ;# proc configure {}

#-------------------------------------------------------------------------------
# Reset all of the character sheet`s information to their default values.
#-------------------------------------------------------------------------------
proc resetAllInfo {} {
   global CSchar CSpow CSdisad CSnote CSinfo CSout

   # reset characteristics info:
   set CSchar(totalCost) {}
   for {set i 0} {$i < $CSchar(count)} {incr i} {
      set CSchar($i,value) {}
      set CSchar($i,cost)  {}
      set CSchar($i,base)  {}
      }
   set CSchar(dataType) {text}

   # reset powers info:
   set CSpow(totalCost) {}
   for {set i 0} {$i < $CSpow(count)} {incr i} {
      set CSpow($i,cost) {}
      set CSpow($i,desc) {}
      set CSpow($i,end)  {}
      }
   set CSpow(dataType) {text}

   # reset disadvantages info:
   set CSdisad(totalCost) {}
   for {set i 0} {$i < $CSdisad(count)} {incr i} {
      set CSdisad($i,cost) {}
      set CSdisad($i,desc) {}
      }
   set CSdisad(dataType) {text}

   # reset notes info:
   for {set i 0} {$i < $CSnote(count)} {incr i} {
      set CSnote($i,desc) {}
      }
   set CSnote(dataType) {text}

   # reset miscellaneous info:
   set CSinfo(name) {}
   set CSinfo(name,type) {text}
   set CSinfo(creator) {}
   set CSinfo(comment) {}
   set CSinfo(imageURL) {}
   set CSinfo(origin,type) {text}

   # clear the contents of the origin text area, if it has been created:
   if [info exists CSinfo(origin,text)] {
      $CSinfo(origin,text) delete 1.0 end
      }

   # clear the contents of the output text area, if it has been created:
   if [info exists CSout(text)] {
      $CSout(text) delete 1.0 end
      }

   }  ;# resetAllInfo {}

#*******************************************************************************
# User interface-building procedures
#*******************************************************************************


#-------------------------------------------------------------------------------
# Build the program`s user interface
#-------------------------------------------------------------------------------
proc buildUserInterface {parent} {
   global CSpage

   # build the frame that will hold the pages:
   set p [frame $parent.pageHolder]
   set CSpage(pageHolder) $p

   # build each page that will go in the pager:
   buildUIPage1 [sectionFrame $p.p1]
   buildUIPage2 [sectionFrame $p.p2]
   buildUIPage3 [sectionFrame $p.p3]
   buildUIPage4 [sectionFrame $p.p4]

   # fill the pager's global array with each page's title and frame:
   set CSpage(count) 4
   set CSpage(0,title) "Characteristics and Information"
   set CSpage(0,page) $p.p1
   set CSpage(1,title) "Powers and Notes"
   set CSpage(1,page) $p.p2
   set CSpage(2,title) "Disadvantages and Origin"
   set CSpage(2,page) $p.p3
   set CSpage(3,title) "Output and Clear"
   set CSpage(3,page) $p.p4

   buildPagerUI $parent CSpage

   }  ;# buildUserInterface {parent}

#-------------------------------------------------------------------------------
# Builds a user interface that allows navigation through the pages in the
# GLOBAL array named $array. The page with index startPage will be 
# displayed initially.
#-------------------------------------------------------------------------------
# page(count)       - the number of pages in the pager   **
# page($i,title)    - the title of the (i+1)th page      **
# page($i,page)     - the top frame for the (i+1)th page **
# page(currIndex)   - the index of the current page
# page(currTitle)   - the title of the current page
# page(currPageNum) - the page number label text for the current page
# page(pageHolder)  - the frame that holds the pages
#
#  ** - these elements must be set before buildPagerUI{} is called
#-------------------------------------------------------------------------------
proc buildPagerUI {parent array {startPage 0}} {
   upvar $array page

   # build the frame that will hold the page title and number, the frame that 
   # will hold the pages, and the frame for the navigation buttons:
   set t [frame $parent.title]
   set p $page(pageHolder)
   set n [frame $parent.nav]

   # build and add the title/page number label:
   label $t.title   -textvar ${array}(currTitle)
   label $t.pageNum -textvar ${array}(currPageNum) -justify right
   pack $t.title   -side left  -anchor w
   pack $t.pageNum -side right -anchor e

   # build and add the navigation buttons:
   button $n.prev -text "Previous" -command "previousPage $array" 
   button $n.next -text "Next" -command "nextPage $array"
   pack $n.prev $n.next -side left -padx 10

   switchToPage $startPage page
   pack $t -side top -fill x
   pack $n -side bottom -pady 5
   pack $p -side top -fill both -expand true

   }  ;# buildPagerUI {parent}

#-------------------------------------------------------------------------------
# Switches to the next page in the pager using the array named $array.
#-------------------------------------------------------------------------------
proc nextPage {array} {
   upvar $array page

   set next [expr ($page(currIndex) + 1) % $page(count)]
   switchToPage $next page
   }

#-------------------------------------------------------------------------------
# Switches to the previous page in the pager using the array named $array.
#-------------------------------------------------------------------------------
proc previousPage {array} {
   upvar $array page

   set prev [expr ($page(currIndex) - 1) % $page(count)]
   switchToPage $prev page
   }

#-------------------------------------------------------------------------------
# Switches to the page with index pageIndex in the pager using the array named
# $array.
#-------------------------------------------------------------------------------
proc switchToPage {newIndex array} {
   upvar $array page

   # unpack the current page (if any) and then pack the new page:
   if [info exists page(currIndex)] {
      pack forget $page($page(currIndex),page)
      }
   pack $page($newIndex,page) -in $page(pageHolder) -fill both -expand true

   # update the information about the (new) current page:
   set page(currIndex) $newIndex
   set page(currTitle) $page($newIndex,title)
   set page(currPageNum) "Page [expr $newIndex + 1]"
   }

#-------------------------------------------------------------------------------
# Build the first page of the user interface (characteristics and names).
#-------------------------------------------------------------------------------
proc buildUIPage1 {parent} {
   global CSchar CSinfo

   set char [frame $parent.char]              ;# characteristics frame
   buildCharacteristicsUI $char CSchar

   set f [frame $parent.f]                    ;# this holds everything else

   set name [sectionFrame $f.name]            ;# names frame + its data type
   buildNamesUI "[frame $name.name]" CSinfo   ;# selection
   buildDataTypeUI "[frame $name.type]" CSinfo(name,type)
   pack $name.name -side top -fill x -expand true -padx 10 -pady 10
   pack $name.type -side top -expand true -padx 10 -pady 10

   # build the frame to hold the characteristics data type selection and a
   # calculator:
   set misc [frame $f.misc]
   buildDataTypeUI "[frame $misc.type]" CSchar(dataType)
   buildCalcUI "[frame $misc.calc]" CScalc
   pack $misc.type -side top -anchor w -padx 20 -pady 20
   pack $misc.calc -side top -anchor w -padx 20 -pady 10 -fill x

   # put the names on the top and the others on the bottom:
   pack $name -side top -fill x -anchor nw
   pack $misc -side top -fill x -anchor nw

   # put the characteristics on the left and the names, etc. on the right:
   pack $char -side left -anchor n -padx 10 -pady 10
   pack $f    -side left -fill both -expand true   
   }

#-------------------------------------------------------------------------------
# Build the second page of the user interface (powers and notes).
#-------------------------------------------------------------------------------
proc buildUIPage2 {parent} {
	global CSpow CSnote

   set pow [frame $parent.pow]                ;# powers frame
   buildPowersUI $pow CSpow

   set f [frame $parent.f]                    ;# this holds everything else

   set note [sectionFrame $f.note]            ;# notes frame + its data type
   label $note.label -text "Notes:"           ;# selection
   buildNotesUI "[frame $note.note]" CSnote
   buildDataTypeUI "[frame $note.type]"  CSnote(dataType)
   pack $note.label -side top -anchor nw -padx 10 -pady 10
   pack $note.note -side top -anchor nw -fill x -expand true -padx 10
   pack $note.type -side top -anchor nw -fill x -expand true -padx 10 -pady 10

   # build the frame to hold the powers data type selection and a calculator:
   set misc [frame $f.misc]
   buildDataTypeUI "[frame $misc.type]" CSpow(dataType)
   buildCalcUI "[frame $misc.calc]" CScalc
   pack $misc.type -side top -anchor w -padx 20 -pady 20
   pack $misc.calc -side top -anchor w -padx 20 -pady 10 -fill x

   # put the notes on the top and the others on the bottom:
   pack $note -side top -anchor nw -fill x
   pack $misc -side top -anchor nw -fill x

   # put the powers on the left and the notes, etc. on the right:
   pack $pow -side left -anchor n -padx 10 -pady 10 -fill x -expand true
   pack $f   -side left -fill both -expand true
   }

#-------------------------------------------------------------------------------
# Build the third page of the user interface (disadvantages and origin).
#-------------------------------------------------------------------------------
proc buildUIPage3 {parent} {
	global CSdisad CSinfo

   set dis  [sectionFrame $parent.disad]
   set orig [sectionFrame $parent.origin]

   # build the disadvantages half of the page:
   buildDataTypeUI "[frame $dis.type]" CSdisad(dataType)
   buildDisadvantagesUI "[frame $dis.disad]" CSdisad
   buildCalcUI "[frame $dis.calc]" CScalc
   pack $dis.type  -side top -anchor nw -fill x -padx 10 -pady 10
   pack $dis.disad -side top -anchor nw -fill x -padx 10
   pack $dis.calc  -side top -anchor nw -fill x -padx 10 -pady 10

   # build the origins half of the page:
   buildDataTypeUI "[frame $orig.type]" CSinfo(origin,type)
   buildOriginUI "[frame $orig.origin]" CSinfo
   pack $orig.type   -side top -anchor nw -fill x -padx 10 -pady 10 
   pack $orig.origin -side top -anchor w -fill both -expand true -padx 10 -pady 10

   # put the disadvantages on the left and the origin on the right:
   pack $dis  -side left -anchor n -fill both -expand true
   pack $orig -side left -fill both -expand true
   }

#-------------------------------------------------------------------------------
# Build the fourth page of the user interface (output and clear).
#-------------------------------------------------------------------------------
proc buildUIPage4 {parent} {
   global CSout

   buildOutputUI [sectionFrame $parent.output] CSout
   buildClearUI  [sectionFrame $parent.clear]  CSout

   pack $parent.clear -side bottom -fill x
   pack $parent.output -side top -expand true -fill both

   }  ;# proc buildUIPage4 {parent}

#-------------------------------------------------------------------------------
# Builds and returns a frame with the specified name that represents a section
# of the user interface.
#-------------------------------------------------------------------------------
proc sectionFrame {name} {
   return [frame $name -bd 2 -relief ridge]   
   }

#-------------------------------------------------------------------------------
# Build the user interface for obtaining the character sheet output
# information. Use the GLOBAL array named $array to store the information.
#-------------------------------------------------------------------------------
proc buildOutputUI {parent array} {
   upvar $array output

   # set the state of the widgets corresponding to each output type (diasbled
   # if it isn`t available, normal if not) and set the initial output type
   # to one that available:
   if $output(browserPolicyAccessed) { # applet w/ Browser security policy access
      set textState "normal"
      set fileState "normal"
      set emailState "normal"
      set browseState "normal"
      set output(type) "email"
      } \
   elseif $output(isApplet) { # applet WITHOUT browser security policy access
      set textState "normal"
      set fileState "disabled"
      set emailState "disabled"
      set browseState "disabled"
      set output(type) "text"
      } \
   else {                     # stand-alone application
      set textState "normal"
      set fileState "normal"
      set emailState "disabled"
      set browseState "disabled"
      set output(type) "text"
      }

   # build a frame to hold the output type UI and the output button:
   set top [frame $parent.top]

   # build a frame to hold the output type UI (question and choices), then
   # build the radiobuttons (and any associated widgets) for each output type:
   set f [frame $top.outType]
   label $f.q -text "Output the character sheet"

   radiobutton $f.text -text "to the text area (below)" -value "text" \
                       -var ${array}(type) -state $textState

   frame $f.file
   radiobutton $f.file.r -text "to the file" -value "file" \
                         -var ${array}(type) -state $fileState
   entry $f.file.e -textvar ${array}(filename) -state $fileState
   pack $f.file.r -side left
   pack $f.file.e -side left -fill x -expand true

   frame $f.email
   radiobutton $f.email.r -text "by e-mail to" -value "email" \
                        -var ${array}(type) -state $emailState
   entry $f.email.e -textvar ${array}(addresses) -state $emailState
   pack $f.email.r -side left
   pack $f.email.e -side left -fill x -expand true

   radiobutton $f.browse -text "to a new browser window" -value "browse" \
                         -var ${array}(type) -state $browseState

   # pack the output type choices widgets:
   pack $f.q      -side top -anchor w -padx 5
   pack $f.text   -side top -anchor w -padx 15
   pack $f.file $f.email -side top -anchor w -fill x -padx 15
   pack $f.browse -side top -anchor w -padx 15

   # build the output button:
   button $top.out -text "Output" -command "onSheetOutput $array"

   # pack output type UI, the output button, and a frame for padding:
   frame $top.pad -width 15
   pack $top.pad $top.out -side right
   pack $f -side left -fill x -expand true -padx 5

   # build the output text area (including scrollbars):
   set t [frame $parent.text]         ;# frame for the whole text area
   frame $t.left                      ;# frame for text area + x scrollbar
   text $t.left.e -bd 2 -relief sunken -width 30 -height 20 \
                  -yscrollcommand "$t.y set" \
                  -xscrollcommand "$t.left.x set"
   set output(text) $t.left.e
   scrollbar $t.left.x -command "$t.left.e xview" -orient horizontal
   scrollbar $t.y -command "$t.left.e yview"

   # pack the scrollbars first so resizing doesn`t hide them:
   pack $t.left.x -side bottom -fill x
   pack $t.left.e -side top    -fill both -expand true
   pack $t.y      -side right  -fill y
   pack $t.left   -side left   -fill both -expand true

   # build the output result area:
   set res [label $parent.res -textvar ${array}(result)]

   # pack the three main parts of the output UI:
   pack $top -side top -fill x
   pack $res -side bottom -anchor w
   pack $t -side top -fill both -expand true

   }  ;# proc buildOutputUI {parent array}

#-------------------------------------------------------------------------------
# Called when the character sheet is to be output.
#-------------------------------------------------------------------------------
proc onSheetOutput {array} {
   upvar $array output

   if [catch {outputSheet} result] {
      set output(result) "Output failed - $result"
      } \
   else {
      set output(result) "Character sheet successfully output"
      }
   }

#-------------------------------------------------------------------------------
# Build the user interface for obtaining information about clearing the
# character sheet. Use the GLOBAL array named $array to store the information.
#-------------------------------------------------------------------------------
proc buildClearUI {parent array} {
   upvar $array clear

   label $parent.lab -text "Clear the entire character sheet?"
   radiobutton $parent.y -text "Yes" -var ${array}(clear) -value 1
   radiobutton $parent.n -text "No"  -var ${array}(clear) -value 0
   button $parent.clear -text "Clear" \
                        -command "doClear ${array}(clear)"
   pack $parent.lab $parent.y $parent.n $parent.clear -side left -padx 5

   set clear(clear) 0

   }  ;# proc buildClearUI {parent array}

#-------------------------------------------------------------------------------
# If clearing the character sheet info has been confirmed, reset the info then
# make sure that clearing is no longer confirmed.
#-------------------------------------------------------------------------------
proc doClear {confirmVarName} {
   upvar $confirmVarName confirmed

   if $confirmed {
      resetAllInfo
      }
   set confirmed 0  ;# clearing is no longer confirmed
   }

#-------------------------------------------------------------------------------
# Builds the user interface for obtaining the characters name, the creator`s
# name and other similar information. Use the GLOBAL array named $array
# to store the names information.
#-------------------------------------------------------------------------------
proc buildNamesUI {parent array} {

   buildNamesLine $parent name     "Character Name:" $array
   buildNamesLine $parent creator  "Creator Name:"   $array
   buildNamesLine $parent comment  "Comment:"        $array
   buildNamesLine $parent imageURL "Image URL:"      $array
   pack $parent.name $parent.creator $parent.comment $parent.imageURL \
        -side top -fill x -expand true
   }

#-------------------------------------------------------------------------------
# Builds one line of the Names user interface with label labelText in the
# parent frame. The entry widget will be tied to the $tag element in the GLOBAL
# array named $array.
#-------------------------------------------------------------------------------
proc buildNamesLine {parent tag labelText array} {

   set f [frame $parent.$tag]
   label $f.label -text $labelText -padx 0
   entry $f.entry -textvar ${array}($tag)
   pack $f.label -side left -anchor w
   pack $f.entry -side left -anchor w -fill x -expand true
   }

#-------------------------------------------------------------------------------
# Build the user interface for obtaining the character`s origins. Use the GLOBAL
# array named $array to store the origin information.
#-------------------------------------------------------------------------------
proc buildOriginUI {parent array} {
   upvar $array origin

   # build the text area (with y-scrollbar):
   text $parent.text -bd 2 -relief sunken -width 30 -height 20 \
                     -yscrollcommand "$parent.yscroll set"
   scrollbar $parent.yscroll -command "$parent.text yview"
   pack $parent.yscroll -side right -fill y
   pack $parent.text    -side left  -fill both -expand true

   set origin(origin,text) $parent.text   ;# the origin text area

   }  ;# proc buildOriginUI {parent array}

#-------------------------------------------------------------------------------
# Build the user interface for obtaining characteristics information. Use the
# GLOBAL array named $array to store the characteristics information.
#-------------------------------------------------------------------------------
proc buildCharacteristicsUI {parent array} {
   upvar $array char

   # set the widths for each type of component:
   set valWidth  4
   set nameWidth 5
   set costWidth 3

   # build and add a row for each characteristic:
   for {set i 0} {$i < $char(count)} {incr i} {
      set f [frame $parent.char$i]
      entry $f.val  -justify right -width $valWidth \
                    -textvar ${array}($i,value)
      label $f.name -width $nameWidth -text $char($i,name)
      entry $f.cost -justify right -width $costWidth \
                    -textvar ${array}($i,cost)
      pack $f.val $f.name $f.cost -side left -anchor n
      pack $f -side top
      }

   # add a total cost row at the bottom of the costs column:
   set f [frame $parent.charTotal]
   entry $f.val  -width $valWidth -state disabled -relief flat
   label $f.name -width $nameWidth -text ""
   entry $f.cost -justify right -width $costWidth -bg grey75 \
                 -textvar ${array}(totalCost)
   pack $f.val $f.name $f.cost -side left
   pack $f -side top

   }  ;# proc buildCharacteristicsUI {parent array}

#-------------------------------------------------------------------------------
# Build the user interface for obtaining powers/skills information. Use the
# GLOBAL array named $array to store the powers information.
#-------------------------------------------------------------------------------
proc buildPowersUI {parent array} {
   upvar $array power

   # set the widths for each type of component:
   set costWidth 4
   set endWidth  3 

   # build and add column labels:
   set f [frame $parent.powHead]
   label $f.cost -text "Cost"
   label $f.desc -text "Powers and Skills"
   label $f.end  -text "END"
   pack $f.cost -side left -anchor w
   pack $f.desc -side left -expand true
   pack $f.end  -side right -anchor e
   pack $f -side top -fill x

   # build and add a row for each power:
   for {set i 0} {$i < $power(count)} {incr i} {
      set f [frame $parent.pow$i]
      entry $f.cost -justify right -width $costWidth \
                    -textvar ${array}($i,cost)
      entry $f.desc -textvar ${array}($i,desc)
      entry $f.end  -justify right -width $endWidth \
                    -textvar ${array}($i,end)
      pack $f.cost -side left
      pack $f.desc -side left -fill x -expand true
      pack $f.end  -side left
      pack $f -side top -fill x -expand true
      }

   # add a total cost row at the bottom of the cost column:
   set f [frame $parent.powTotal]
   entry $f.cost -justify right -width $costWidth -bg grey75 \
                 -textvar ${array}(totalCost)
   pack $f.cost -side left 
   pack $f -side top -fill x -expand true

   }  ;# proc buildPowersUI {parent array}

#-------------------------------------------------------------------------------
# Build the user interface for obtaining disadvantages information. Use the
# GLOBAL array named $array to store the disadvantages information.
#-------------------------------------------------------------------------------
proc buildDisadvantagesUI {parent array} {
   upvar $array disad

   # set the width of the cost entry components:
   set costWidth 3

   # build and add column labels:
   set f [frame $parent.disadHead]
   label $f.cost -text "Pts."
   label $f.desc -text "Disadvantages"
   pack $f.cost -side left -anchor w
   pack $f.desc -side left -expand true
   pack $f -side top -fill x

   # build and add a row for each disadvantage:
   for {set i 0} {$i < $disad(count)} {incr i} {
      set f [frame $parent.disad$i]
      entry $f.cost -justify right -width $costWidth \
                    -textvar ${array}($i,cost)
      entry $f.desc -textvar ${array}($i,desc)
      pack $f.cost -side left
      pack $f.desc -side left -fill x -expand true
      pack $f -side top -fill x -expand true
      }

   # add a total cost row at the bottom of the cost column:
   set f [frame $parent.disadTotal]
   entry $f.cost -justify right -width $costWidth -bg grey75 \
                 -textvar ${array}(totalCost)
   pack $f.cost -side left -anchor w 
   pack $f -side top -fill x -expand true

   }  ;# proc buildDisadvantagesUI {parent array}

#-------------------------------------------------------------------------------
# Build a user interface for getting notes. Use the GLOBAL array named
# $array to store the notes information.
#-------------------------------------------------------------------------------
proc buildNotesUI {parent array} {
   upvar $array note

   # build the notes entry fields:
   for {set i 0} {$i < $note(count)} {incr i} {
      entry $parent.$i -textvar ${array}($i,desc)
      pack $parent.$i -side top -anchor w -fill x -expand true
      }
   }

#-------------------------------------------------------------------------------
# Builds a user interface to get the type of the information to be entered in
# a given section. Puts this information in the global variable $typeVar.
#-------------------------------------------------------------------------------
proc buildDataTypeUI {parent typeVar} {

   label $parent.txt1 -bd 0 -pady 0 -text "Treat all of the information"
   label $parent.txt2 -bd 0 -pady 0 -text "entered in this section as"
   
   set f [frame $parent.radio]
   radiobutton $f.text -text "Text" -var $typeVar -value {text} \
                                    -selectcolor #aabbcc
   radiobutton $f.html -text "HTML" -var $typeVar -value {html} \
                                    -selectcolor #ff4400
   pack $f.text $f.html -side left -padx 0

   pack $parent.txt1 $parent.txt2 $f -side top
   }

#-------------------------------------------------------------------------------
# Builds a calculator user interface in the parent frame. Use the GLOBAL array
# named $array to store the calculator`s information.
#-------------------------------------------------------------------------------
proc buildCalcUI {parent array} {
   global tcl_precision
   upvar $array calc

   # build the calculator UI: the question entry box, the equals sign label
   # and the answer label
   set f $parent
   entry $f.e -textvar ${array}($f,q) -bg #aabbcc
   label $f.s -text "=" -padx 2
   label $f.l -textvar ${array}($f,a) \
              -width [expr {$tcl_precision + 5}] -bg #aabbcc

   # pressing Return performs the calculation; selecting the question entry
   # box selects its entire contents:
   bind $f.e <Return>  "doCalc $f $array; $f.e select range 0 end"
   bind $f.e <FocusIn> "$f.e select range 0 end"

   $f.e insert 0 "Calculator"     ;# initial contents of the calculator

   # pack the answer label first so that it isn`t clipped when the applet
   # is too small:
   pack $f.l $f.s -side right
   pack $f.e -side left  -fill x -expand true

   }  ;# proc buildCalcUI {parent array}

#-------------------------------------------------------------------------------
# Updates the calculator identified by ID by performing its calculation. The
# required information is stored in the array named $array.
#-------------------------------------------------------------------------------
proc doCalc {ID array} {
   upvar $array calc
   set q $calc($ID,q)

   # do nothing iff there is no calculation specified:
   if {[string compare $q ""] == 0} {return}

   # set the answer to the result of the calculation - unless an error occurs,
   # in which case set the answer to 'ERR':
   if [catch {expr $q} calc($ID,a)] {set calc($ID,a) "ERR"}
   }

#*******************************************************************************
# Character sheet output procedures
#*******************************************************************************


#-------------------------------------------------------------------------------
# Converts the contents of a table cell to HTML given its dataType. Returns the
# converted text, or an error if dataType is an unknown data type.
#-------------------------------------------------------------------------------
proc cellToHTML {cellContents dataType} {

   # no matter the data type, if the cell is empty return a line break:
   if {[string compare "[string trim $cellContents]" ""] == 0} {return "<br>"}

   return [lineToHTML $cellContents $dataType]
   }

#-------------------------------------------------------------------------------
# Converts a line of text to HTML given its dataType. Returns the converted
# text, or an error if dataType is an unknown data type.
#-------------------------------------------------------------------------------
proc lineToHTML {line dataType} {
   switch -- $dataType {
      html {return $line}
      text {return [convertEntities $line]}
      default {error "Unknown data type: $dataType"}
      }
   }

#-------------------------------------------------------------------------------
# Converts an arbitrary piece of text to HTML given its dataType. Returns the
# converted text, or an error if dataType is an unknown data type.
#-------------------------------------------------------------------------------
proc textToHTML {text dataType} {
   switch -- $dataType {
      html {return $text}
      text {
         # convert entities and remove any leading/trailing newlines:
         set text [string trim [convertEntities $text] "\n"]

         # start a new paragraph after each newline:
         regsub -all -- "\n" $text "\n<p>" text
         return $text
         }
      default {error "Unknown data type: $dataType"}
      }
   }

#-------------------------------------------------------------------------------
# Converts special HTML characters in text to their entity representations.
# Returns the modified text.
#-------------------------------------------------------------------------------
proc convertEntities {text} {
   regsub -all -- {&} $text {\&amp;} text
   regsub -all -- {<} $text {\&lt;}  text
   regsub -all -- {>} $text {\&gt;}  text
   return $text
   }

#-------------------------------------------------------------------------------
# Output the character sheet in HTML. Returns an error (from beforeSheetOutput)
# if the sheet is not successfully output.
#-------------------------------------------------------------------------------
proc outputSheet {} {
   global tab CSinfo
   set tab "   "

   beforeSheetOutput
   outputHTMLPrologue

   # output the creator`s name as an HTML comment iff it was specified:
   if {[string compare $CSinfo(creator) ""] != 0} {
      outputString "\n<!-- Created by: $CSinfo(creator) -->\n"
      }

   # output the comment as an HTML comment iff it was specified:
   if {[string compare $CSinfo(comment) ""] != 0} {
      outputString "\n<!-- $CSinfo(comment) -->\n"
      }

   outputName
   outputCharacteristics left
   outputPowers right "*"
   outputString "<br clear=all>\n"

   outputDisadvantages left
   outputNotes right 
   outputString "<br clear=all>\n"

   outputPicture right
   outputOrigin

   outputHTMLEpilogue
   afterSheetOutput
   
   }  ;# proc outputSheet {}

#-------------------------------------------------------------------------------
# Called before a character sheet is to be output. Throws an error if the
# character sheet cannot be output.
#-------------------------------------------------------------------------------
proc beforeSheetOutput {} {
   global CSout

   switch -- $CSout(type) {
      text    {$CSout(text) delete 1.0 end}
      email   {set CSout(buffer) ""}
      browse  {browser_openFrame _blank}
      file    {set CSout(fileID) [open $CSout(filename) {WRONLY CREAT TRUNC}]}
      default {error "Unknown output type $CSout(type)"}
      }
   }

#-------------------------------------------------------------------------------
# Outputs the specified string as part of the generated character sheet.
#-------------------------------------------------------------------------------
proc outputString {str} {
   global CSout

   switch -- $CSout(type) {
      text    {$CSout(text) insert end $str}
      email   {append CSout(buffer) $str}
      browse  {browser_writeFrame _blank $str}
      file    {puts $CSout(fileID) $str}
      default {error "Unknown output type $CSout(type)"}
      }
   }

#-------------------------------------------------------------------------------
# Called after the entire character sheet has been output.
#-------------------------------------------------------------------------------
proc afterSheetOutput {} {
   global CSout

   switch -- $CSout(type) {
      text    { }
      email   {sendEMail $CSout(buffer) $CSout(addresses)}
      browse  {browser_closeFrame _blank}
      file    {catch [close $CSout(fileID)]}
      default {error "Unknown output type $CSout(type)"}
      }
   }

#-------------------------------------------------------------------------------
# Sends msg as e-mail to the (comma-separated) list of e-mail addresses. Assumes
# that if it is called then the program is allowed to send e-mail.
#-------------------------------------------------------------------------------
proc sendEMail {msg addresses} {
   browser_sendEMail $addresses $msg
   }

#-------------------------------------------------------------------------------
# Outputs the character`s name.
#-------------------------------------------------------------------------------
proc outputName {} {
   global CSinfo

   set name [lineToHTML $CSinfo(name) $CSinfo(name,type)]
   if {[string compare $name ""] != 0} {
      outputString "\n<h1>${name}</h1>\n"
      }
   }

#-------------------------------------------------------------------------------
# Outputs the character`s picture with the specified alignment and width.
#-------------------------------------------------------------------------------
proc outputPicture {align {width {}}} {
   global CSinfo

   set URL [string trim $CSinfo(imageURL)]
   if {[string compare $URL ""] != 0} {
      set str "\n<img src=\"${URL}\" align=${align}"
      if {[string compare $width ""] != 0} {
         append str " width=${width}"
         }
      outputString "${str}>\n"
      }
   }

#-------------------------------------------------------------------------------
# Outputs the characteristics portion of the character sheet with the
# specified alignment and width.
#-------------------------------------------------------------------------------
proc outputCharacteristics {align {width {}}} {
   global CSchar tab

   set dt $CSchar(dataType)              ;# data type of characteristics

   # build the head of the table:
   set comment "characteristics"
   set str "\n<!-- $comment -->\n"
   append str "<table border=1 cellspacing=1 cellpadding=2 align=${align}"
   if {[string compare $width ""] != 0} {
      append str " width=${width}"
      }
   append str ">\n"
   append str [tableHeadingsRow {text} "Val" "Char" "Pts"]

   # build a row for each characteristic:
   for {set i 0} {$i < $CSchar(count)} {incr i} {
      set val  [cellToHTML $CSchar($i,value) $dt]
      set name [cellToHTML $CSchar($i,name)  $dt]
      set cost [cellToHTML $CSchar($i,cost)  $dt]

      append str "${tab}<tr>\n" \
                 "${tab}${tab}<td align=right>${val}</td> " \
                 "<td>${name}</td> <td align=right>${cost}</td>\n" \
                 "${tab}</tr>\n"
      }

   # build a row for the total characteristics cost:
   set total [cellToHTML $CSchar(totalCost) $dt]
   append str "${tab}<tr>\n" \
              "${tab}${tab}<td colspan=2 align=right><b>Char Cost =</b></td> " \
              "<td align=right><b>${total}</b></td>\n" \
              "${tab}</tr>\n"

   # build the foot of the table:
   append str "</table>  <!-- $comment -->\n"

   outputString $str

   }  ;# proc outputCharacteristics {}

#-------------------------------------------------------------------------------
# Outputs the powers portion of the character sheet with the specified
# alignment and width.
#-------------------------------------------------------------------------------
proc outputPowers {align {width {}}} {
   global CSpow tab

   set dt $CSpow(dataType)              ;# data type of powers

   # build the head of the table:
   set comment "powers, skills, etc."
   set str "\n<!-- $comment -->\n"
   append str "<table border=1 cellspacing=1 cellpadding=2 align=${align}"
   if {[string compare $width ""] != 0} {
      append str " width=${width}"
      }
   append str ">\n"
   append str [tableHeadingsRow {text} "Pts" "Powers and Skills" "END"]

   # build a row for each power:
   set last [lastPowerIndex]
   for {set i 0} {$i <= $last} {incr i} {
      set cost [cellToHTML $CSpow($i,cost) $dt]
      set desc [cellToHTML $CSpow($i,desc) $dt]
      set end  [cellToHTML $CSpow($i,end)  $dt]

      append str "${tab}<tr>\n" \
                 "${tab}${tab}<td align=right valign=top>${cost}</td>\n" \
                 "${tab}${tab}<td>${desc}</td>\n" \
                 "${tab}${tab}<td align=right valign=top>${end}</td>\n" \
                 "${tab}</tr>\n"
      }

   # build a row for the total powers cost:
   set total [cellToHTML $CSpow(totalCost) $dt]
   append str "${tab}<tr>\n" \
              "${tab}${tab}<td align=right><b>${total}</b></td> " \
              "<td colspan=2><b>= Powers Cost</b></td>\n" \
              "${tab}</tr>\n"

   # build the foot of the table:
   append str "</table>  <!-- $comment -->\n"

   outputString $str

   }  ;# proc outputPowers {}

#-------------------------------------------------------------------------------
# Outputs the disadvantages portion of the character sheet with the specified
# alignment and width.
#-------------------------------------------------------------------------------
proc outputDisadvantages {align {width {}}} {
   global CSdisad tab

   set dt $CSdisad(dataType)             ;# data type of disadvantages

   # build the head of the table:
   set comment "disadvantages"
   set str "\n<!-- $comment -->\n"
   append str "<table border=1 cellspacing=1 cellpadding=2 align=${align}"
   if {[string compare $width ""] != 0} {
      append str " width=${width}"
      }
   append str ">\n"

   append str [tableHeadingsRow {text} "Pts" "Disadvantages"]

   # build a row for each disadvantage:
   set last [lastDisadIndex]
   for {set i 0} {$i <= $last} {incr i} {
      set cost [cellToHTML $CSdisad($i,cost) $dt]
      set desc [cellToHTML $CSdisad($i,desc) $dt]

      append str "${tab}<tr>\n" \
                 "${tab}${tab}<td align=right valign=top>${cost}</td>\n" \
                 "${tab}${tab}<td>${desc}</td>\n" \
                 "${tab}</tr>\n"
      }

   # build a row for the total disadvantages cost/value:
   set total [cellToHTML $CSdisad(totalCost) $dt]
   append str "${tab}<tr>\n" \
              "${tab}${tab}<td align=right><b>${total}</b></td> " \
              "<td><b>= Disadvantages Total</b></td>\n" \
              "${tab}</tr>\n"

   # build the foot of the table:
   append str "</table>  <!-- $comment -->\n"

   outputString $str

   }  ;# proc outputDisadvantages {}

#-------------------------------------------------------------------------------
# Outputs the notes portion of the character sheet with the specified alignment 
# and width. If there are no notes then nothing is output.
#-------------------------------------------------------------------------------
proc outputNotes {align {width {}}} {
   global CSnote tab

   # if there are no notes the don`t output a notes section at all:
   set last [lastNoteIndex]
   if {$last == -1} {return}

   set dt $CSnote(dataType)              ;# data type of notes

   # build the head of the table:
   set comment "notes"
   set str "\n<!-- $comment -->\n"
   append str "<table border=1 cellspacing=1 cellpadding=2 align=${align}"
   if {[string compare $width ""] != 0} {
      append str " width=${width}"
      }
   append str ">\n"
   append str [tableHeadingsRow {text} "Notes"]

   # build a row for each note:
   for {set i 0} {$i <= $last} {incr i} {
      set desc [cellToHTML $CSnote($i,desc) $dt]

      append str "${tab}<tr>\n" \
                 "${tab}${tab}<td>${desc}</td>\n" \
                 "${tab}</tr>\n"
      }

   # build the foot of the table:
   append str "</table>  <!-- $comment -->\n"

   outputString $str

   }  ;# proc outputNotes {}

#-------------------------------------------------------------------------------
# Output the origin of the character.
#-------------------------------------------------------------------------------
proc outputOrigin {} {
   global CSinfo tab

   set str "\n<!-- origin -->\n"

   # set text to the contents of the text area holding the origin:
   set endIndex [$CSinfo(origin,text) index "end -1 chars"]
   set text [textToHTML "[$CSinfo(origin,text) get 1.0 $endIndex]" \
                        $CSinfo(origin,type)]
   if {[string compare "[string trim $text]" ""] != 0} {
      append str "<p>${text}</p>\n"
      }

   outputString $str
   }

#-------------------------------------------------------------------------------
# Returns the index of the last power whose fields are not all whitespace, or
# -1 if there is no such index.
#-------------------------------------------------------------------------------
proc lastPowerIndex {} {
   global CSpow

   for {set i [expr {$CSpow(count) - 1}]} {$i >= 0} {incr i -1} {
      set all $CSpow($i,cost)
      append all $CSpow($i,desc) $CSpow($i,end)
      if {[string compare "[string trim $all]" ""] != 0} {return $i}
      }
   return -1 
   }

#-------------------------------------------------------------------------------
# Returns the index of the last disadvantage whose fields are not all 
# whitespace, or -1 if there is no such index.
#-------------------------------------------------------------------------------
proc lastDisadIndex {} {
   global CSdisad

   for {set i [expr {$CSdisad(count) - 1}]} {$i >= 0} {incr i -1} {
      set all $CSdisad($i,cost)
      append all $CSdisad($i,desc)
      if {[string compare "[string trim $all]" ""] != 0} {return $i}
      }
   return -1
   }

#-------------------------------------------------------------------------------
# Returns the index of the last note whose fields are not all 
# whitespace, or -1 if there is no such index.
#-------------------------------------------------------------------------------
proc lastNoteIndex {} {
   global CSnote

   for {set i [expr {$CSnote(count) - 1}]} {$i >= 0} {incr i -1} {
      set all $CSnote($i,desc)
      if {[string compare "[string trim $all]" ""] != 0} {return $i}
      }
   return -1
   }

#-------------------------------------------------------------------------------
# Returns an HTML table row of the specified headings.
#-------------------------------------------------------------------------------
proc tableHeadingsRow {dataType args} {
   global tab

   set str "${tab}<tr>\n${tab}${tab}"
   foreach h $args {
      append str "<th>[cellToHTML $h $dataType]</th> "
      }
   append str "\n${tab}</tr>\n"

   return $str
   }

#-------------------------------------------------------------------------------
# Outputs the standard prologue of an HTML document generated by this program.
#-------------------------------------------------------------------------------
proc outputHTMLPrologue {} {
   global tab CSinfo

   set name [lineToHTML $CSinfo(name) $CSinfo(name,type)]
   set str "<!DOCTYPE HTML PUBLIC \"-//IETF//DTD HTML 3.2//EN\">\n"
   append str "<html>\n\n<head>\n" \
              "${tab}<title>${name}</title>\n" \
              "${tab}<meta name=\"GENERATOR\" content=\"CharForm 1.0\">\n" \
              "</head>\n\n<body>\n"
   outputString $str
   }

#-------------------------------------------------------------------------------
# Outputs the standard epilogue of an HTML document.
#-------------------------------------------------------------------------------
proc outputHTMLEpilogue {} {
   outputString "\n</body>\n</html>\n"
   }

#-------------------------------------------------------------------------------
# main program
#-------------------------------------------------------------------------------

# determine whether this is a standalone app or an applet and
# whether we can access the browser security policy:
set CSout(isApplet) [array exists embed_args]  ;# is there a better test ???
set CSout(browserPolicyAccessed) [expr [catch {package require Browser}] ? 0 : 1]

configure
buildUserInterface {}
